#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

#include "bonnie.h"
#include "fileio.h"

CPCCHAR rand_chars = "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

CFileOp::~CFileOp()
{
  close();
  if(m_name)
  {
    int len = strlen(m_name);
    for(int i = 0; i < m_num_files; i++)
    {
      sprintf(&m_name[len], ".%02d", i);
      unlink(m_name);
    }
    delete m_name;
  }
}

int CFileOp::seek(int offset, int whence)
{
  switch(whence)
  {
    case SEEK_SET:
      m_file_ind = offset / m_chunks_per_file;
      m_cur_pos = offset % m_chunks_per_file;
    break;
    case SEEK_CUR:
      m_cur_pos += offset;
      while(m_cur_pos < 0)
      {
        m_file_ind--;
        m_cur_pos += m_chunks_per_file;
      }
      while(m_cur_pos >= m_chunks_per_file)
      {
        m_file_ind++;
        m_cur_pos -= m_chunks_per_file;
      }
    break;
    default:
      fprintf(stderr, "unsupported seek option\n");
      return -1;
  }
  if(m_file_ind < 0 || m_file_ind >= m_num_files
     || (m_file_ind == m_num_files - 1 && m_cur_pos >= m_last_file_chunks))
  {
    fprintf(stderr, "Bad seek offset\n");
    return -1;
  }
  int rc = lseek(m_fd[m_file_ind], m_cur_pos << ChunkBits, SEEK_SET);
  if(rc == -1)
    fprintf(stderr, "Error in lseek to %d\n", (m_cur_pos << ChunkBits));
  else
    rc = 0;
  return rc;
}

int CFileOp::read_block(PVOID buf)
{
  int rc = ::read(m_fd[m_file_ind], buf, Chunk);
  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    seek(0, SEEK_CUR);
  }
  return rc;
}

int CFileOp::read_block_getc(char *buf)
{
  int next;
  for(int i = 0; i < Chunk; i++)
  {
    if ((next = getc(m_stream[m_file_ind])) == EOF)
    {
      fprintf(stderr, "Can't getc(3)\n");
      return -1;
    }
    /* just to fool optimizers */
    buf[next]++;
  }

  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    seek(0, SEEK_CUR);
    if(fseek(m_stream[m_file_ind], m_cur_pos * Chunk, SEEK_SET) == -1)
    {
      fprintf(stderr, "Can't fseek()\n");
      return -1;
    }
  }
  return 0;
}

int CFileOp::write_block(CPCVOID buf)
{
  int rc = ::write(m_fd[m_file_ind], buf, Chunk);
  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    seek(0, SEEK_CUR);
  }
  return rc;
}

int CFileOp::write_block_putc()
{
  for(int i = 0; i < Chunk; i++)
  {
    if (putc(i & 0x7f, m_stream[m_file_ind]) == EOF)
    {
      fprintf(stderr, "Can't putc() - disk full?\n");
      return -1;
    }
  }
  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    seek(0, SEEK_CUR);
    if(fseek(m_stream[m_file_ind], m_cur_pos * Chunk, SEEK_SET) == -1)
    {
      fprintf(stderr, "Can't fseek()\n");
      return -1;
    }
  }
  return 0;
}

int CFileOp::open(CPCCHAR basename, bool create)
{
  m_name = new char[strlen(basename) + 4];
  strcpy(m_name, basename);
  return reopen(create);
}

CFileOp::CFileOp()
 : m_max_file_size(1024)
 , m_chunks_per_file(Unit / Chunk * m_max_file_size)
 , m_isopen(false), m_name(NULL)
{
}

void CFileOp::set_size(int size)
{
  m_total_chunks = Unit / Chunk * size;
  m_last_file_chunks = m_total_chunks % m_chunks_per_file;
  m_num_files = size / 1024 + 1;
}

typedef FILE * PFILE;

int CFileOp::reopen(bool create)
{
  int i;
  m_cur_pos = 0;
  m_file_ind = 0;

  if(m_isopen) close();

  m_isopen = true;
  m_stream = new PFILE[m_num_files];
  m_fd = new int[m_num_files];
  for(i = 0; i < m_num_files; i++)
  {
    m_fd[i] = -1;
    m_stream[i] = NULL;
  }
  int len = strlen(m_name);
  for(i = 0; i < m_num_files; i++)
  {
    sprintf(&m_name[len], ".%02d", i);
    if(m_open(m_name, i, create))
      return true;
  }
  m_name[len] = '\0';
  return 0;
}

int CFileOp::m_open(CPCCHAR basename, int ind, bool create)
{
  if(create)
  { /* create from scratch */
    unlink(basename);
    m_fd[ind] = ::open(basename, O_RDWR | O_CREAT | O_EXCL, 0777);
  } /* create from scratch */
  else
    m_fd[ind] = ::open(basename, O_RDWR, 0777);

  if (m_fd[ind] == -1)
  {
    fprintf(stderr, "Can't open file %s\n", basename);
    return -1;
  }
  m_stream[ind] = fdopen(m_fd[ind], "r+");
  if (m_stream[ind] == NULL)
  {
    fprintf(stderr, "Cant fdopen()\n");
    return -1;
  }
  return 0;
}

void CFileOp::close()
{
  if(!m_isopen)
    return;
  for(int i = 0; i < m_num_files; i++)
  {
    if(m_stream[i]) fclose(m_stream[i]);
    if(m_fd[i]) ::close(m_fd[i]);
  }
  m_isopen = false;
  delete m_fd;
  delete m_stream;
}


/*
 * Do a typical-of-something random I/O.  Any serious application that
 *  has a random I/O bottleneck is going to be smart enough to operate
 *  in a page mode, and not stupidly pull individual words out at
 *  odd offsets.  To keep the cache from getting too clever, some
 *  pages must be updated.  However an application that updated each of
 *  many random pages that it looked at is hard to imagine.  
 * However, it would be wrong to put the update percentage in as a
 *  parameter - the effect is too nonlinear.  Need a profile
 *  of what Oracle or Ingres or some such actually does.
 * Be warned - there is a *sharp* elbow in this curve - on a 1-Mb file,
 *  most substantial unix systems show >2000 random I/Os per second -
 *  obviously they've cached the whole thing and are just doing buffer
 *  copies.  
 */
int 
CFileOp::doseek(long where, bool update)
{
  int   buf[Chunk / IntSize];
  int   size;

  if (seek(where, SEEK_SET) == -1)
    return io_error("lseek in doseek");
  if ((size = read_block(PVOID(buf))) == -1)
    return io_error("read in doseek");

  /* every so often, update a block */
  if (update)
  { /* update this block */

    /* touch a word */
    buf[((int) random() % (size/IntSize - 2)) + 1]--;
    if (seek(where, SEEK_SET) == -1)
      return io_error("lseek in doseek update");
    if (write_block(PVOID(buf)) == -1)
      return io_error("write in doseek");
  } /* update this block */
  return 0;
}

COpenTest::COpenTest()
{
  m_number = 0;
  m_used = NULL;
  m_dirname = NULL;
  m_file_name_buf = NULL;
  m_file_names = NULL;
}

COpenTest::~COpenTest()
{
  if(m_used)
  {
    fprintf(stderr, "Cleaning up test directory after error.\n");
    for(int i = 0; i < m_number; i++)
    {
      unlink(m_file_names[i]);
    }
    chdir("..");
    delete m_used;
    m_used = NULL;
  }
  if(m_dirname)
  {
    rmdir(m_dirname);
    delete m_dirname;
  }
  delete m_file_name_buf;
  delete m_file_names;
}

void COpenTest::make_names(int num, bool do_random)
{
  m_number = num;
  delete m_file_name_buf;
  delete m_file_names;
  m_file_name_buf = new char[(MaxNameLen + 1) * num];
  m_file_names = new PCHAR[num];
  PCHAR buf = m_file_name_buf;
  for(int i = 0; i < num; i++)
  {
    char rand_buf[RandExtraLen + 1];
    int len = random() % (RandExtraLen + 1);
    int j;
    for(j = 0; j < len; j++)
    {
      rand_buf[j] = rand_chars[random() % strlen(rand_chars)];
    }
    rand_buf[j] = '\0';
    if(do_random)
    {
      sprintf(buf, "%s%7d", rand_buf, i);
    }
    else
    {
      sprintf(buf, "%7d%s", i, rand_buf);
    }
    m_file_names[i] = buf;
    buf += strlen(buf) + 1;
  }
}

int COpenTest::create(CPCCHAR dirname, BonTimer &timer, int num, bool do_random)
{
  make_names(num, do_random);
  timer.timestamp();
  m_number = num;
  m_dirname = new char[strlen(dirname) + 1];
  strcpy(m_dirname, dirname);

  if(mkdir(dirname, S_IRWXU))
  {
    fprintf(stderr, "Can't make directory %s\n", dirname);
    return -1;
  }
  if(chdir(dirname))
  {
    fprintf(stderr, "Can't change to directory %s\n", dirname);
    return -1;
  }
  m_used = new char[num];
  int i;
  for(i = 0; i < m_number; i++)
    m_used[i] = false;

  for(i = 0; i < m_number; i++)
  {
    int fd = creat(m_file_names[i], S_IRUSR | S_IWUSR);
    if(fd == -1)
    {
      fprintf(stderr, "Can't create file %s\n", m_file_names[i]);
      return -1;
    }
    close(fd);
    m_used[i] = true;
  }
  timer.get_delta_t(do_random ? CreateRand : CreateSeq);
  return 0;
}

int COpenTest::delete_random(BonTimer &timer)
{
  timer.timestamp();
  int i;
  for(i = 0; i < m_number; i++)
  {
    int j;
    for(j = random() % m_number; !m_used[j]; j++)
    {
      if(j == m_number - 1)
        j = -1;
    }
    // j is now the index of an unused entry
    if(unlink(m_file_names[j]))
    {
      fprintf(stderr, "Can't delete file %s\n", m_file_names[j]);
      return -1;
    }
    m_used[j] = false;
  }
  delete m_used;
  m_used = NULL;
  chdir("..");
  rmdir(m_dirname);
  delete m_dirname;
  m_dirname = NULL;
  timer.get_delta_t(DelRand);
  return 0;
}

int COpenTest::delete_sequential(BonTimer &timer)
{
  timer.timestamp();
  DIR *d = opendir(".");
  if(!d)
  {
    fprintf(stderr, "Can't open directory.\n");
    return -1;
  }
  dirent *file_ent;

  int count = 0;
  while((file_ent = readdir(d)) != NULL)
  {
    if(file_ent->d_name[0] != '.')
    {
      if(unlink(file_ent->d_name))
      {
        fprintf(stderr, "Can't delete file %s\n", file_ent->d_name);
        return -1;
      }
      count++;
    }
  }
  delete m_used;
  m_used = NULL;
  chdir("..");
  rmdir(m_dirname);
  delete m_dirname;
  m_dirname = NULL;
  closedir(d);
  if(count != m_number)
  {
    fprintf(stderr, "Expected %d files but only got %d\n", m_number, count);
    return -1;
  }
  timer.get_delta_t(DelSeq);
  return 0;
}

int COpenTest::stat_random(BonTimer &timer)
{
  timer.timestamp();

  int i;
  for(i = 0; i < m_number; i++)
  {
    struct stat buf;
    char *name = m_file_names[random() % m_number];
    if(stat(name, &buf))
    {
      fprintf(stderr, "Can't stat file %s\n", name);
      return -1;
    }
  }
  timer.get_delta_t(StatRand);
  return 0;
}

int COpenTest::stat_sequential(BonTimer &timer)
{
  timer.timestamp();
  DIR *d = opendir(".");
  if(!d)
  {
    fprintf(stderr, "Can't open directory.\n");
    return -1;
  }
  dirent *file_ent;
  int count = 0;
  while((file_ent = readdir(d)) != NULL)
  {
    struct stat buf;
    if(file_ent->d_name[0] != '.') // our files do not start with a dot
    {
      if(stat(file_ent->d_name, &buf))
      {
        fprintf(stderr, "Can't stat file %s\n", file_ent->d_name);
        return -1;
      }
      count++;
    }
  }
  closedir(d);
  if(count != m_number)
  {
    fprintf(stderr, "Expected %d files but only got %d\n", m_number, count);
    return -1;
  }
  timer.get_delta_t(StatSeq);
  return 0;
}

